"""
Array of intervals corresponding to IntervalArray of Pandas.
Used for IntervalIndex, which is necessary for Series.value_counts() with 'bins'
argument.
"""
import numba
import pandas as pd
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo


class IntervalType(types.Type):

    def __init__(self):
        super(IntervalType, self).__init__('IntervalType()')


class IntervalArrayType(types.ArrayCompatible):

    def __init__(self, arr_type):
        self.arr_type = arr_type
        self.dtype = IntervalType()
        super(IntervalArrayType, self).__init__(name=
            f'IntervalArrayType({arr_type})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    def copy(self):
        return IntervalArrayType(self.arr_type)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(IntervalArrayType)
class IntervalArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        jakbd__cgri = [('left', fe_type.arr_type), ('right', fe_type.arr_type)]
        models.StructModel.__init__(self, dmm, fe_type, jakbd__cgri)


make_attribute_wrapper(IntervalArrayType, 'left', '_left')
make_attribute_wrapper(IntervalArrayType, 'right', '_right')


@typeof_impl.register(pd.arrays.IntervalArray)
def typeof_interval_array(val, c):
    arr_type = bodo.typeof(val._left)
    return IntervalArrayType(arr_type)


@intrinsic
def init_interval_array(typingctx, left, right=None):
    assert left == right, 'Interval left/right array types should be the same'

    def codegen(context, builder, signature, args):
        mrte__qzan, syc__lowo = args
        ndlu__zipio = cgutils.create_struct_proxy(signature.return_type)(
            context, builder)
        ndlu__zipio.left = mrte__qzan
        ndlu__zipio.right = syc__lowo
        context.nrt.incref(builder, signature.args[0], mrte__qzan)
        context.nrt.incref(builder, signature.args[1], syc__lowo)
        return ndlu__zipio._getvalue()
    aaggb__eqr = IntervalArrayType(left)
    gfg__bwic = aaggb__eqr(left, right)
    return gfg__bwic, codegen


def init_interval_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    xro__ihch = []
    for tfvq__skgv in args:
        rel__xsf = equiv_set.get_shape(tfvq__skgv)
        if rel__xsf is not None:
            xro__ihch.append(rel__xsf[0])
    if len(xro__ihch) > 1:
        equiv_set.insert_equiv(*xro__ihch)
    left = args[0]
    if equiv_set.has_shape(left):
        return ArrayAnalysis.AnalyzeResult(shape=left, pre=[])
    return None


(ArrayAnalysis._analyze_op_call_bodo_libs_interval_arr_ext_init_interval_array
    ) = init_interval_array_equiv


def alias_ext_init_interval_array(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 2
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)
    numba.core.ir_utils._add_alias(lhs_name, args[1].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_interval_array',
    'bodo.libs.int_arr_ext'] = alias_ext_init_interval_array


@box(IntervalArrayType)
def box_interval_arr(typ, val, c):
    ndlu__zipio = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, typ.arr_type, ndlu__zipio.left)
    olkpz__xdmq = c.pyapi.from_native_value(typ.arr_type, ndlu__zipio.left,
        c.env_manager)
    c.context.nrt.incref(c.builder, typ.arr_type, ndlu__zipio.right)
    ask__lna = c.pyapi.from_native_value(typ.arr_type, ndlu__zipio.right, c
        .env_manager)
    dndlp__yguu = c.context.insert_const_string(c.builder.module, 'pandas')
    deem__lfl = c.pyapi.import_module_noblock(dndlp__yguu)
    yjrsc__ipj = c.pyapi.object_getattr_string(deem__lfl, 'arrays')
    dlw__gougi = c.pyapi.object_getattr_string(yjrsc__ipj, 'IntervalArray')
    mbhr__xxkpz = c.pyapi.call_method(dlw__gougi, 'from_arrays', (
        olkpz__xdmq, ask__lna))
    c.pyapi.decref(olkpz__xdmq)
    c.pyapi.decref(ask__lna)
    c.pyapi.decref(deem__lfl)
    c.pyapi.decref(yjrsc__ipj)
    c.pyapi.decref(dlw__gougi)
    c.context.nrt.decref(c.builder, typ, val)
    return mbhr__xxkpz


@unbox(IntervalArrayType)
def unbox_interval_arr(typ, val, c):
    olkpz__xdmq = c.pyapi.object_getattr_string(val, '_left')
    left = c.pyapi.to_native_value(typ.arr_type, olkpz__xdmq).value
    c.pyapi.decref(olkpz__xdmq)
    ask__lna = c.pyapi.object_getattr_string(val, '_right')
    right = c.pyapi.to_native_value(typ.arr_type, ask__lna).value
    c.pyapi.decref(ask__lna)
    ndlu__zipio = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    ndlu__zipio.left = left
    ndlu__zipio.right = right
    lndb__cju = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(ndlu__zipio._getvalue(), is_error=lndb__cju)


@overload(len, no_unliteral=True)
def overload_interval_arr_len(A):
    if isinstance(A, IntervalArrayType):
        return lambda A: len(A._left)


@overload_attribute(IntervalArrayType, 'shape')
def overload_interval_arr_shape(A):
    return lambda A: (len(A._left),)


@overload_attribute(IntervalArrayType, 'ndim')
def overload_interval_arr_ndim(A):
    return lambda A: 1


@overload_attribute(IntervalArrayType, 'nbytes')
def overload_interval_arr_nbytes(A):
    return lambda A: A._left.nbytes + A._right.nbytes


@overload_method(IntervalArrayType, 'copy', no_unliteral=True)
def overload_interval_arr_copy(A):
    return lambda A: bodo.libs.interval_arr_ext.init_interval_array(A._left
        .copy(), A._right.copy())
